package com.css.simulation.resource.scheduler.service;

import api.common.pojo.constants.DictConstants;
import api.common.util.*;
import com.css.simulation.resource.scheduler.manager.TaskIndexManager;
import com.css.simulation.resource.scheduler.manager.TaskManager;
import com.css.simulation.resource.scheduler.mapper.IndexTemplateMapper;
import com.css.simulation.resource.scheduler.mapper.ManualProjectMapper;
import com.css.simulation.resource.scheduler.mapper.TaskIndexMapper;
import com.css.simulation.resource.scheduler.mapper.TaskMapper;
import com.css.simulation.resource.scheduler.pojo.po.IndexTemplatePO;
import com.css.simulation.resource.scheduler.pojo.po.ProjectPO;
import com.css.simulation.resource.scheduler.pojo.po.TaskIndexPO;
import com.css.simulation.resource.scheduler.pojo.po.TaskPO;
import com.css.simulation.resource.scheduler.pojo.to.ScoreTO;
import com.css.simulation.resource.scheduler.util.MinioUtil;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.minio.MinioClient;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.sshd.client.SshClient;
import org.apache.sshd.client.session.ClientSession;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.util.*;
import java.util.concurrent.atomic.AtomicReference;
import java.util.stream.Collectors;

@Service
@Slf4j
public class TaskService {

    private final String USER_ID = "simulation-resource-scheduler";


    @Autowired
    CloseableHttpClient closeableHttpClient;
    @Autowired
    RequestConfig requestConfig;
    @Autowired
    MinioClient minioClient;
    @Value("${minio.bucket-name}")
    String bucketName;
    @Autowired
    StringRedisTemplate redisTemplate;
    @Autowired
    ManualProjectMapper manualProjectMapper;
    //    @Autowired
//    TaskManager taskManager;
    @Autowired
    TaskMapper taskMapper;
    @Autowired
    TaskIndexManager taskIndexManager;
    @Autowired
    TaskManager taskManager;
    @Autowired
    TaskIndexMapper taskIndexMapper;
    @Autowired
    IndexTemplateMapper indexTemplateMapper;
    @Value("${scheduler.manual-project.topic}")
    String manualProjectTopic;
    @Value("${scheduler.manual-project.result-path-minio}")
    String resultPathMinio;
    @Value("${scheduler.score.hostname}")
    String hostnameScore;
    @Value("${scheduler.score.username}")
    String usernameScore;
    @Value("${scheduler.score.password}")
    String passwordScore;

    @Value("${spring.kafka.hostname}")
    String hostnameKafka;
    @Value("${spring.kafka.username}")
    String usernameKafka;
    @Value("${spring.kafka.password}")
    String passwordKafka;
    @Value("${spring.kafka.delete-command}")
    String kafkaDeleteCommand;
    @Value("${scheduler.score.py-path}")
    String pyPath;
    @Value("${scheduler.linux-temp-path}")
    String linuxTempPath;
    @Value("${simulation-cloud.client-id}")
    String clientId;
    @Value("${simulation-cloud.client-secret}")
    String clientSecret;
    @Value("${simulation-cloud.token-uri}")
    String tokenUri;
    @Value("${simulation-cloud.evaluation-level-uri}")
    String evaluationLevelUri;


    @SneakyThrows
    public void taskState(String taskId, String state, String podName) {

        String projectId = taskMapper.selectProjectIdById(taskId);
        redisTemplate.opsForValue().set(manualProjectTopic + ":" + projectId + ":" + taskId + ":pod", podName);
        String podDeleteCommand = "kubectl delete pod " + podName;
        SshClient client = SshUtil.getClient();
        ClientSession session = SshUtil.getSession(client, hostnameScore, usernameScore, passwordScore);
        if ("Running".equals(state)) {
            log.info("TaskService--state 修改任务 " + taskId + "的状态为 Running，pod 名称为：" + podName);
            taskMapper.updateStateWithStartTime(taskId, state, TimeUtil.getNowForMysql());
        } else if ("Aborted".equals(state)) {
            log.info("TaskService--state 修改任务 " + taskId + "的状态为 Aborted，pod 名称为：" + podName
                    + "，并执行删除 pod 命令：" + podDeleteCommand);
            SshUtil.execute(session, podDeleteCommand);
//            taskManager.updateFailStateWithStopTime(taskId, state, TimeUtil.getNowForMysql()); // 如果任务 abort 代表项目失败
            //result-path-minio: /project/manual-project/
            String minioPathOfErrorLog = resultPathMinio + projectId + "/" + taskId + "error.log";
            boolean objectExist = MinioUtil.isObjectExist(minioClient, bucketName, minioPathOfErrorLog);
            String targetEvaluate;
            if (objectExist) {
                String errorString = MinioUtil.downloadToString(minioClient, bucketName, minioPathOfErrorLog);
                String[] lines = errorString.split("\n");
                StringBuilder errorMessage = new StringBuilder();
                for (String line : lines) {
                    if (line.startsWith("Original Error")) {
                        errorMessage.append(line).append("\n");
                    }

                    if (line.startsWith("Possible Cause")) {
                        errorMessage.append(line);
                        break;
                    }
                }
                targetEvaluate = errorMessage.toString();
            } else {
                targetEvaluate = DictConstants.TASK_ERROR_REASON_2;
            }


            taskMapper.updateFailStateWithStopTime(taskId, state, TimeUtil.getNowForMysql(), targetEvaluate);
        } else if ("Terminated".equals(state)) {
            log.info("TaskService--state 修改任务 " + taskId + "的状态为 Terminated，pod 名称为：" + podName
                    + "，并执行删除 pod 命令：" + podDeleteCommand);
            SshUtil.execute(session, podDeleteCommand);
            taskMapper.updateFailStateWithStopTime(taskId, state, TimeUtil.getNowForMysql(), DictConstants.TASK_ERROR_REASON_3);
        } else if ("PendingAnalysis".equals(state)) {
            log.info("TaskService--state 修改任务 " + taskId + "的状态为 PendingAnalysis，pod 名称为：" + podName
                    + "，并执行删除 pod 命令：" + podDeleteCommand);
            SshUtil.execute(session, podDeleteCommand);
            taskMapper.updateSuccessStateWithStopTime(taskId, state, TimeUtil.getNowForMysql());
        } else {
            log.error("TaskService--state 出现了未知状态:" + state);
            taskMapper.updateFailStateWithStopTime(taskId, state, TimeUtil.getNowForMysql(), DictConstants.TASK_ERROR_REASON_4);
        }
        ProjectPO projectPO = manualProjectMapper.selectById(taskId);
        if (projectPO == null) {
            session.close();
            client.stop();
            return;
        }
        Set<String> keys = redisTemplate.keys("manualProject:" + projectId + "*");
        assert keys != null;
        redisTemplate.delete(keys);
        String scenePackageId = projectPO.getScenePackageId();  // 场景测试包 id，指标根 id
//        log.info("------- /state 任务 " + taskId + " 的父项目为：" + projectId);
        int taskNum = manualProjectMapper.selectTaskNumById(projectId);
        int endTaskNum = manualProjectMapper.selectEndTaskNum(projectId);    // 查询已结束的任务 'Aborted', 'PendingAnalysis', 'Terminated'
        manualProjectMapper.updateTaskCompleted(projectId, endTaskNum);
        log.info("TaskService--state 项目 " + projectId + " 完成进度为：" + endTaskNum + "/" + taskNum);
        if (taskNum != endTaskNum) {  // 已结束任务数等于所有任务数量，才会准备打分；否则退出。
            session.close();
            client.stop();
            return;
        }

        log.info("结束项目的 job");
        SshUtil.execute(session, "kubectl delete job project-" + projectId);

        // 删除 kafka topic
//        SshClient clientKafka = SshUtil.getClient();
//        ClientSession sessionKafka = SshUtil.getSession(clientKafka, hostnameKafka, usernameKafka, passwordKafka);
//        String topicDeleteCommand = StringUtil.replace(kafkaDeleteCommand, "topicName", projectId);
//        SshUtil.execute(sessionKafka, topicDeleteCommand);
//        SshUtil.stop(clientKafka, sessionKafka);

        List<TaskPO> taskList = taskMapper.selectTaskListByProjectId(projectId);  // 所有任务信息
        int taskNumber = taskList.size();
        log.info("TaskService--state 共有 " + taskNumber + "个任务！");
        taskIndexMapper.deleteFirstByProjectId(projectId);
        taskIndexMapper.deleteLastByProjectId(projectId);
        // -------------------------------- 查询叶子指标 --------------------------------
        List<IndexTemplatePO> leafIndexTemplateList = indexTemplateMapper.selectLeafIndexWithRuleDetailsByPackageId(scenePackageId);
        List<TaskIndexPO> leafTaskIndexList = new ArrayList<>();
        log.info("TaskService--state 共有 " + leafIndexTemplateList.size() + "个叶子节点！");

//        SshClient clientScore = SshUtil.getClient();
//        ClientSession sessionScore = SshUtil.getSession(clientScore, hostnameScore, usernameScore, passwordScore);
        for (int i = 0; i < leafIndexTemplateList.size(); i++) {
            AtomicReference<String> scoreExplain = new AtomicReference<>(); // 每个叶子指标下的任务的得分说明一样和叶子指标一致
            IndexTemplatePO indexTemplatePO = leafIndexTemplateList.get(i);
            String indexId = indexTemplatePO.getIndexId();
            String parentId = indexTemplatePO.getParentId();
            String rootId = indexTemplatePO.getRootId();
            String weight = indexTemplatePO.getWeight();
            log.info("TaskService--state 开始执行对第 " + (i + 1) + " 个叶子节点 " + indexId + " 进行打分！");
            String ruleName = indexTemplatePO.getRuleName();    // 打分脚本名称，例如 AEB_1-1
            String ruleDetails = indexTemplatePO.getRuleDetails();    // 打分脚本内容
            String ruleFilePath = pyPath + "scripts/" + ruleName.split("_")[0] + "/" + ruleName + ".py";
            log.info("TaskService--state 将叶子节点 " + indexId + " 对应的打分规则保存到临时目录：" + ruleFilePath);
            FileUtil.writeInputStreamToLocalFile(IoUtil.stringToInputStream(ruleDetails), ruleFilePath);

            List<TaskPO> taskListOfLeafIndex = taskList.stream()
                    .filter(task -> indexId.equals(task.getLastTargetId()))
                    .collect(Collectors.toList());

            log.info("TaskService--state 叶子节点 " + indexId + " 包括 " + taskListOfLeafIndex.size() + " 个成功运行结束任务！");
            log.info("TaskService--state 计算叶子节点 " + indexId + " 的得分：" + taskListOfLeafIndex);
            taskListOfLeafIndex.forEach(task2 -> {
                String runState = task2.getRunState();
                if (DictConstants.TASK_ANALYSIS.equals(runState)) {
                    String task2Id = task2.getId();
                    taskMapper.updateSuccessStateWithStopTime(task2Id, DictConstants.TASK_ANALYSING, TimeUtil.getNowForMysql());
                    // 计算每个任务的得分
                    String runResultMinio = task2.getRunResultFilePath() + "/Ego.csv";
                    String runResultLinux = linuxTempPath + runResultMinio;

//                        python3 /home/ubuntu/test/Evaluate/main.py /home/ubuntu/test/test_data.csv 4 AEB_1-2
//                        String command = "python3 " + pyPath + " " + runResultLinux + " " + task2.getSceneType();  // 默认使用场景名称找打分脚本
                    String scoreCommand = "python3 " + pyPath + "main.py " + runResultLinux + " " + task2.getSceneType() + " " + ruleName; // 指定打分脚本
                    String scoreResult;
                    try {
                        try {
                            log.info("TaskService--state 下载 minio 上的结果文件 " + runResultMinio + " 到本地：" + runResultLinux);
                            MinioUtil.downloadToFile(minioClient, bucketName, runResultMinio, runResultLinux);  // 也可改成下载到指定ip的服务器上，需要保证和打分脚本在一台机器上。
                        } catch (Exception e) {
                            throw new RuntimeException("------- TaskService--state 下载 minio 上的结果文件出错：" + e.getMessage());
                        }
                        try {
                            log.info("TaskService--state 开始执行打分命令：" + scoreCommand);
//                            scoreResult = SshUtil.execute(sessionScore, command);
                            scoreResult = SshUtil.execute(session, scoreCommand);
                            log.info("TaskService--state 项目" + projectId + "的任务" + task2Id + "打分结束，结果为：" + scoreResult);
                        } catch (IOException e) {
                            throw new RuntimeException("------- TaskService--state 项目" + projectId + "的任务" + task2Id + " 打分出错，命令为：" + scoreCommand + " 修改状态为：" + DictConstants.TASK_ABORTED + "\n" + e.getMessage());
                        }
                    } catch (Exception e) {
                        task2.setRunState(DictConstants.TASK_ABORTED);
                        taskMapper.updateFailStateWithStopTime(task2Id, DictConstants.TASK_ABORTED, TimeUtil.getNowForMysql(), DictConstants.TASK_ERROR_REASON_5);
                        throw new RuntimeException(e.getMessage());
                    }
                    ScoreTO score = null;
                    try {
                        String replace = StringUtil.replace(scoreResult, "'", "\"");
                        score = JsonUtil.jsonToBean(replace, ScoreTO.class);
                    } catch (JsonProcessingException e) {
                        e.printStackTrace();
                    }
                    assert score != null;
                    task2.setReturnSceneId(score.getUnit_scene_ID());
//                    task2.setScore(new Random().nextInt(10) * 10.0);
                    task2.setScore(score.getUnit_scene_score());
                    task2.setTargetEvaluate(score.getEvaluate_item());
                    task2.setScoreExplain(score.getScore_description());
                    task2.setModifyUserId(USER_ID);
                    task2.setModifyTime(TimeUtil.getNowForMysql());
                    scoreExplain.set(score.getScore_description());

                    task2.setRunState(DictConstants.TASK_COMPLETED);
                    taskMapper.updateSuccessStateAndScoreResultWithStopTime(
                            task2,
                            DictConstants.TASK_COMPLETED,
                            TimeUtil.getNowForMysql()
                    );
                }
            });


            // 计算不合格的任务数（不到100分就是不合格，执行失败的不算）
            long notStandardSceneNumber = taskListOfLeafIndex.stream()
                    .filter(task -> task.getScore() < 100)
                    .count();

            // 计算总分
            double leafSum = taskListOfLeafIndex.stream()
                    .mapToDouble(TaskPO::getScore)
                    .sum();
            // 计算任务的个数
            long resultNumberOfCurrentIndex = taskListOfLeafIndex.size();
            log.info("TaskService--state 项目 " + projectId + " 的叶子指标" + indexId + "下成功执行的场景数量为：" + resultNumberOfCurrentIndex);
            double leafIndexScore = resultNumberOfCurrentIndex == 0 ? 0 : NumberUtil.cut(leafSum / resultNumberOfCurrentIndex, 2);
            // -------------------------------- 保存叶子指标得分 --------------------------------
            indexTemplatePO.setTempScore(leafIndexScore);
            TaskIndexPO leafTaskIndex = TaskIndexPO.builder()
                    .id(StringUtil.getRandomUUID())
                    .pId(projectId)
                    .target(indexTemplatePO.getIndexId())
                    .notStandardSceneNum((int) notStandardSceneNumber)
                    .score(leafIndexScore)
                    .indexId(indexId)
                    .parentId(parentId)
                    .rootId(rootId)
                    .weight(weight)
                    .scoreExplain(scoreExplain.get())
                    .build();
            leafTaskIndex.setCreateUserId(USER_ID);
            leafTaskIndex.setCreateTime(TimeUtil.getNowForMysql());
            leafTaskIndex.setModifyUserId(USER_ID);
            leafTaskIndex.setModifyTime(TimeUtil.getNowForMysql());
            leafTaskIndex.setIsDeleted("0");
            leafTaskIndexList.add(leafTaskIndex);
        }
//        SshUtil.stop(clientScore, sessionScore);
//        // 保存任务分数
//        taskManager.batchUpdateByScoreResult(taskList);
        // 保存末级指标分数
        taskIndexManager.batchInsertLeafIndex(leafTaskIndexList);
        // 保存一级指标分数
        log.info("TaskService--state 项目 " + projectId + " 的所有任务分数为：" + taskList);
        log.info("TaskService--state 根据每个指标的得分和权重算出各个一级指标的得分（即 project 对应的场景测试包下的一级指标）！");
        computeFirst(leafTaskIndexList, projectId);

        // 调用 server 的接口，计算评价等级
        String tokenUrl = tokenUri + "?grant_type=client_credentials"
                + "&client_id=" + clientId
                + "&client_secret=" + clientSecret;
        log.info("TaskService--state 获取仿真云平台 token：" + tokenUrl);
        String response = HttpUtil.get(closeableHttpClient, requestConfig, tokenUrl);
        ObjectMapper objectMapper = new ObjectMapper();
        JsonNode jsonNode = objectMapper.readTree(response);
        String accessToken = jsonNode.path("access_token").asText();
        log.info("TaskService--state 仿真云平台 token 为：" + accessToken);
        Map<String, String> headers = new HashMap<>();
        headers.put("Authorization", "Bearer " + accessToken);
        Map<String, String> params = new HashMap<>();
        params.put("id", projectId);
        String post = HttpUtil.post(closeableHttpClient, requestConfig, evaluationLevelUri, headers, params);
        log.info("TaskService--state 访问仿真云平台评价等级接口：" + evaluationLevelUri + "，请求头为：" + headers + "，请求体为：" + params + "结果为：" + post);
        log.info("TaskService--state 项目 " + projectId + " 打分完成！");
        manualProjectMapper.updateProjectState(projectId, DictConstants.PROJECT_COMPLETED, TimeUtil.getNowForMysql());   // 修改该 project 的状态为已完成
        log.info("TaskService--state 项目 " + projectId + " 执行完成！");
        session.close();
        client.stop();

    }


    public Boolean taskConfirm(String taskId) {
        // 查询 task 如果不是 pending 则不执行
        String state = taskMapper.selectStateById(taskId);
        return DictConstants.TASK_PENDING.equals(state);
    }


    public void computeFirst(List<TaskIndexPO> leafTaskIndexList, String projectId) {

        log.info("------- /state computeFirst 计算父指标得分：" + leafTaskIndexList);
        Iterator<TaskIndexPO> leafTaskIndexIterator = leafTaskIndexList.iterator();
        while (leafTaskIndexIterator.hasNext()) {
            TaskIndexPO leafTaskIndex = leafTaskIndexIterator.next();
            String parentId = leafTaskIndex.getParentId();
            String rootId = leafTaskIndex.getRootId();
            if (parentId.equals(rootId)) {
                leafTaskIndex.setCreateUserId(USER_ID);
                leafTaskIndex.setCreateTime(TimeUtil.getNowForMysql());
                leafTaskIndex.setModifyUserId(USER_ID);
                leafTaskIndex.setModifyTime(TimeUtil.getNowForMysql());
                leafTaskIndex.setIsDeleted("0");
                taskIndexMapper.insertFirstIndex(leafTaskIndex);
                leafTaskIndexIterator.remove();
            }

        }
        if (leafTaskIndexList.size() > 0) {
            Map<String, List<TaskIndexPO>> sonTaskIndexMap = leafTaskIndexList.stream().collect(Collectors.groupingBy(TaskIndexPO::getParentId));
            Set<String> parentIdSet = sonTaskIndexMap.keySet();
            List<String> parentIdList = CollectionUtil.setToList(parentIdSet);
            List<IndexTemplatePO> parentIndexTemplateList = indexTemplateMapper.selectByIdList(parentIdList);
            // 计算父指标得分
            List<TaskIndexPO> parentIndexList = new ArrayList<>();
            parentIndexTemplateList.forEach(indexTemplate -> {
                String weight = indexTemplate.getWeight();

                List<TaskIndexPO> sonTaskIndexList = sonTaskIndexMap.get(indexTemplate.getIndexId());
                double parentScore = sonTaskIndexList.stream().mapToDouble(taskIndex -> taskIndex.getScore() * Double.parseDouble(taskIndex.getWeight()) / 100).sum();
                TaskIndexPO parentTaskIndex = TaskIndexPO.builder()
                        .id(StringUtil.getRandomUUID())
                        .pId(projectId)
                        .target(indexTemplate.getIndexId())
                        .score(((int) parentScore * 100) / 100.0)
                        .indexId(indexTemplate.getIndexId())
                        .parentId(indexTemplate.getParentId())
                        .rootId(indexTemplate.getRootId())
                        .weight(weight)
                        .build();
                parentIndexList.add(parentTaskIndex);
            });
            // 将父指标作为叶子指标递归
            computeFirst(parentIndexList, projectId);
        }
    }

    public void taskTick(String taskId) {
        log.info("TaskService--taskTick 任务 " + taskId + "心跳！");
        // 刷新 redis 心跳时间
        ProjectPO projectPO = manualProjectMapper.selectById(taskId);
        String projectId = projectPO.getId();
        redisTemplate.opsForValue().set(manualProjectTopic + ":" + projectId + ":" + taskId, TimeUtil.getNowString());
    }

}



